@@ -1,6 +1,6 @@
 #![allow(clippy::mutable_key_type)]
 use std::{
-    collections::BTreeMap,
+    collections::{BTreeMap, HashSet},
     mem::ManuallyDrop,
     ops::{Bound, RangeBounds},
     sync::{
@@ -132,10 +132,10 @@ impl<'a> Transaction<'a> {
     /// Gets a key/value pair.
     pub fn get(&self, key: &[u8]) -> Result<Option<Bytes>> {
         if let Some(res) = self.cache.get(key) {
-            res.pseudotime.store(
-                GLOBAL_PSEUDO_TIME.fetch_add(1, Ordering::Relaxed),
-                Ordering::Relaxed,
-            );
+            if res.deleted {
+                return Ok(None);
+            }
+            res.update_pseudotime();
             let value = res.value.clone();
             Ok(Some(value))
         } else {
@@ -185,6 +185,7 @@ impl<'a> Transaction<'a> {
 struct CacheEntry {
     value: Bytes,
     pseudotime: AtomicU64,
+    deleted: bool,
 }
 
 impl CacheEntry {
@@ -192,8 +193,20 @@ impl CacheEntry {
         Self {
             value,
             pseudotime: AtomicU64::new(GLOBAL_PSEUDO_TIME.fetch_add(1, Ordering::Relaxed)),
+            deleted: false,
         }
     }
+
+    fn update_pseudotime(&self) {
+        self.pseudotime.store(
+            GLOBAL_PSEUDO_TIME.fetch_add(1, Ordering::Relaxed),
+            Ordering::Relaxed,
+        )
+    }
+
+    fn get_pseudotime(&self) -> u64 {
+        self.pseudotime.load(Ordering::Relaxed)
+    }
 }
 
 // Global pseudotime counter
@@ -250,7 +263,7 @@ impl DictInner {
 
     /// Maybe garbage collect
     fn maybe_gc(&self) -> Result<()> {
-        if nanorand::tls_rng().generate_range(0u32, 10000) == 0 {
+        if nanorand::tls_rng().generate_range(0u32, 100000) == 0 {
             let mut cache = self.cache.write();
 
             // we flush everything
@@ -258,15 +271,23 @@ impl DictInner {
             assert_eq!(0, self.send_change.len());
 
             if cache.len() > self.gc_threshold {
-                log::debug!("garbage collect started!");
+                log::warn!("garbage collect started!");
                 let old_len = cache.len();
-                // TODO: LRU rather than randomly killing
+                // approximate LRU
+                let highest_time = GLOBAL_PSEUDO_TIME.load(Ordering::Relaxed);
+                let oldest_time = highest_time.saturating_sub((old_len / 2) as u64);
                 let mut new = BTreeMap::new();
+                let mut seen_pseudotimes = HashSet::new();
                 for (k, v) in cache.iter() {
-                    new.insert(k.clone(), CacheEntry::new(v.value.clone()));
+                    if !seen_pseudotimes.insert(v.get_pseudotime()) {
+                        panic!("duplicate pseudotime");
+                    }
+                    if !v.deleted && v.get_pseudotime() > oldest_time {
+                        new.insert(k.clone(), CacheEntry::new(v.value.clone()));
+                    }
                 }
                 *cache = new;
-                log::debug!("GC complete: {} => {}", old_len, cache.len());
+                log::warn!("GC complete: {} => {}", old_len, cache.len());
             }
         }
         Ok(())
@@ -284,9 +305,15 @@ impl DictInner {
     /// Deletes a key.
     fn remove(&self, key: &[u8]) -> Result<Option<Bytes>> {
         let mut cache = self.cache.write();
-        let previous = match cache.remove(key) {
+        let previous = match cache.get_mut(key) {
             None => self.read_uncached(&key)?,
-            Some(val) => Some(val.value),
+            Some(val) => {
+                if val.deleted {
+                    return Ok(None);
+                }
+                val.deleted = true;
+                Some(val.value.clone())
+            }
         };
         // we now signal the background thread
         self.send_change
@@ -304,13 +331,7 @@ impl DictInner {
         // if there's a previous value in the cache, perfect! that's our previous value
         // otherwise, the previous value is in the disk. the disk value cannot possibly be out of date, because if there are inflight writes, there would be cache.
         let previous = cache
-            .insert(
-                key.clone(),
-                CacheEntry {
-                    value: value.clone(),
-                    pseudotime: AtomicU64::new(GLOBAL_PSEUDO_TIME.fetch_add(1, Ordering::Relaxed)),
-                },
-            )
+            .insert(key.clone(), CacheEntry::new(value.clone()))
             .map(|v| v.value);
         // we now signal the background thread
         let actual_previous = match previous {
@@ -326,25 +347,29 @@ impl DictInner {
     /// Reads, using the cache
     fn read(&self, key: &[u8]) -> Result<Option<Bytes>> {
         self.maybe_gc()?;
-        let cache = self.cache.upgradable_read();
-        // first we check the cache
-        if let Some(res) = cache.get(key) {
-            res.pseudotime.store(
-                GLOBAL_PSEUDO_TIME.fetch_add(1, Ordering::Relaxed),
-                Ordering::Relaxed,
-            );
-            let value = res.value.clone();
-            Ok(Some(value))
-        } else {
-            let value = self.read_uncached(key)?;
-            // try to upgrade to write lock. if we can't, that's okay too
-            let cache = RwLockUpgradableReadGuard::try_upgrade(cache);
-            if let Ok(mut cache) = cache {
-                if let Some(value) = value.as_ref() {
-                    cache.insert(Bytes::copy_from_slice(key), CacheEntry::new(value.clone()));
+        loop {
+            let cache = self.cache.upgradable_read();
+            // first we check the cache
+            if let Some(res) = cache.get(key) {
+                if res.deleted {
+                    return Ok(None);
                 }
+                res.update_pseudotime();
+                let value = res.value.clone();
+                return Ok(Some(value));
+            } else {
+                let value = self.read_uncached(key)?;
+                // try to upgrade to write lock. if we can't, that's okay too
+                let cache = RwLockUpgradableReadGuard::try_upgrade(cache);
+                if let Ok(mut cache) = cache {
+                    if let Some(value) = value.as_ref() {
+                        cache.insert(Bytes::copy_from_slice(key), CacheEntry::new(value.clone()));
+                    }
+                } else {
+                    continue;
+                }
+                return Ok(value);
             }
-            Ok(value)
         }
     }
 
@@ -472,6 +497,9 @@ fn sync_to_disk(
             instructions.len(),
             max_batch_size,
         );
+
+        // continue;
+
         if instructions.len() >= max_batch_size {
             max_batch_size += 10;
         }
